// 	KEDS Project                                                     CodeIndex.c																											(Kansas Event Data System)
//__________________________________________________________________________________
// 																																	 CodeIndex.c
/*

This program produces an index from a TABARI .actors or .verbs dictionary.

USER INPUT

The program asks for 

1. Name of the file to be indexed

2. Whether this is an actor (A) or verb file (V)

3. Whether the complete input string should be stored (Y), or just the pattern (N)

4. Whether the output file should be an HTML table (Y), or tabbed text (N)

Error checking is *not* done in steps (2) - (4)

INPUT FILES

The input files are standard TABARI .actors or .verbs dictionaries.

SCREEN FEEDBACK

Program  shows the entry that is currently being evaluated -- this is useful
in the event that the program crashes due to some unexpected input, but
slows the program slightly.

OUTPUT FILES

The output has the name of the input file followed by the suffix ".idx.html"
or ".idx.html". Actor codes are indented when they match on 3-character blocks 
(following the CAMEO actor coding conventions), and verb (event) codes on the 
third and fourth character (following WEIS and CAMEO conventions)

NOTES

1. There is reasonable error checking for situations where there could be 
   memory problems.
   
2. The code-length conventions are hard-coded as "magic numbers" in the
   routines WriteIndex and WriteTable; these really should be constants
	 
SYSTEM REQUIREMENTS:
This program was originally developed on the Macintosh operating system (OS 10.x)
but contains no Macintosh-specific features and will presumably work on any
system that has a gcc compiler. Current version was compiled under Macintosh
OS-10.4

To compile: gcc CodeIndex.c
To run:     ./a.out

Compiler: gcc
Language: ANSI C
For plausible indenting of this source code, set the tab size in your editor to "2"

REVISION HISTORY
  July-05:  Initial version (0.8B1), based very loosely on OneADayFilter.c

PROVENANCE
Programmer: Philip A. Schrodt
            Kansas Event Data System Project
            Center for International Political Analysis
            University of Kansas 
            Lawrence, Kansas, 66044 U.S.A.
            http://www.ku.edu/~keds
			email: schrodt@ku.edu
						
This program was funded by US National Science Foundation grant SES-0455158.

Copyright (c) 2005  Philip A. Schrodt.  All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted under the terms of the GNU General Public License:
http://www.opensource.org/licenses/gpl-license.html

Report bugs to: schrodt@ku.edu

*/

//__________________________________________________________________________________

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

//__________________________________________________________________________________
// 	Macros

#define	FALSE			0
#define TRUE			1

//#define IN_FILE		"APSA96.actors" // Debugging
//#define IN_FILE		"Balkans.actors.updated" // Debugging
#define IN_FILE		"STANDARD.VERBS.FINAL" // Debugging
#define MAX_FILE_NAME	48			// maximum length of file name 
#define TXT_SUFFIX	".idx.txt"		// suffix for tabbed text output file
#define HTML_SUFFIX	".idx.html"		// suffix for the HTML output file

//#define MAXSTORE  2000				// maximum size of record storage array	
#define MAX_INPUT_LINE   256	// maximum size of event record
#define MAX_CODE_REC_LEN   128	// maximum size of composite code record
#define MAX_CODE_LEN   16				// maximum size of individual code record
#define MAX_TEXT   524288	// maximum length (plus 1) of text storage
#define WORDWRAP   32		// maximum line length in HTML table


//__________________________________________________________
// Structures

struct codeListStruct { 	 // event list elements
	char code[MAX_CODE_LEN];				//  event code
	struct textListStruct * ptext;  // next list item
	struct codeListStruct * pnext;  // next list item
}* codeListHead;;

struct textListStruct { 	 // target list elements
	int	 itext;						// frequency
	struct textListStruct * pnext;  // next list item
};

//___________________________________________________________________________
// Prototypes

void Initialize();  					// initialize globals
void GetFileNames();  				// get names of files from keyboard
int  GetRecord(void);					// read record from fin; returns false if feof(fin)
int  AddText(void);  				// adds sinput to record information; returns index of start of text
int  AddVerb(void);  				// adds sinput to record information; returns index of start of text
void WriteIndex(void); // display list as tabbed text
void WriteTable(void); // display list as HTML table
void AddTextRec(struct codeListStruct * pcLS); // add icurtext index to pcLS
void AddRecord(void); // Adds record to index
void Report_Error(char * s);  // exit on error

//__________________________________________________________
// Globals

FILE	*fin, *ffilt, *fdisc;				// input and output files
char  sfin_name[MAX_FILE_NAME], sfilt_name[MAX_FILE_NAME]; // names of same
//int		doFull = TRUE;			// store full text?
int		doFull = FALSE;			// store full text?
int		isActor = FALSE;			// processing actor file?
int		doHTML  = TRUE;			// write HTML table
//int		doHTML  = FALSE;			// write HTML table
  
char  textBuf[MAX_TEXT]; // storage for text
int   itextBuf = 0;	// index to same
int   icurtext = 0;	// start of most recently added text
char  src[MAX_CODE_LEN] = "XXX";		// source code 
char  sinput[MAX_INPUT_LINE];			// read input line
char  scode[MAX_CODE_REC_LEN] = "";		// extracted code
char  sverb[MAX_CODE_REC_LEN] = "";		// current verb
char  c;

//____________________________________________________________________________
//																																main program

int main(void)
{
char c;
int ka = 0;
char * sp;
char * spr;

Initialize();
GetFileNames();

// while ((ka++<480) && GetRecord())  // debugging
while (GetRecord())  
{ 
  printf("Input %d: %scode: %s\n",ka,sinput, scode);
  if (!scode[0]) continue; // no codes found
  if (isActor) icurtext = AddText();
  else  icurtext = AddVerb();
  sp = scode;
  while (*sp) {  // copy next code into src
  	spr = src;
  	while ('/' != *sp) *spr++ = *sp++;
  	++sp; // skip over /
  	*spr++ = '\0';
    AddRecord();
 }
}
if (doHTML) WriteTable();
else WriteIndex();
putchar('\n');
fclose(fin);
fclose(ffilt);
printf("\aFinished!\n");
	
} // main

void Initialize()
// initialize globals
{
// initialize list head
	codeListHead =  malloc(sizeof(struct codeListStruct));
	if (NULL == codeListHead) Report_Error("Insufficient memory to initialize codeListHead"); // of course, if we ever hit this we're in big trouble...
	strcpy(codeListHead->code,"   ");		
	strcpy(sinput,"Incorrectly formed codes\n");
  icurtext = AddText();
	codeListHead->ptext = malloc(sizeof(struct textListStruct));1;
	codeListHead->ptext->itext = icurtext;
	codeListHead->ptext->pnext = NULL;
	codeListHead->pnext = NULL;
	strcpy(src,"###");		
	strcpy(sinput,"Complex codes:");
  icurtext = AddText();
  AddRecord();
	strcpy(src,"---");		
	strcpy(sinput,"Discard codes:");
  icurtext = AddText();
  AddRecord();
	strcpy(sinput,"[no file inputinitialization: no file input]");
} // Initialize

void GetFileNames()  
// get name and characteristics of file from keyboard
{
char c;

#if TRUE
	printf("\n=========================================================================\n\n");
	printf("CodeIndex.c: Program for producing indices of TABARI dictionaries\n");
	printf("              Version 0.8B1\n");
	printf("              Kansas Event Data System (KEDS)\n");
	printf("\n(c) 2005, University of Kansas\n");
	printf("Redistribution and use in source and binary forms, with or without modification,\n");
	printf("are permitted under the terms of the GNU General Public License:\n");
	printf("http://www.opensource.org/licenses/gpl-license.html\n\n");
    printf("This program was funded by US National Science Foundation grant SES-0455158.\n");
#endif

#if TRUE					// get names from keyboard
	printf("\aEnter name of file to be indexed:");
	scanf("%s",sfin_name);
	if (strlen(sfin_name) >= MAX_FILE_NAME - 8)  Report_Error("Input file name is too long to allow \".idx\" suffix to be added ");
	
	printf("Is this an actor (A) or verb (V) file? (A/V):");
	scanf("%c%c",&sinput[1],&sinput[0]);  // first %c traps the \n from previous input
	if (('v' == sinput[0]) || ('V' == sinput[0])) isActor = FALSE;
	else isActor = TRUE;

	printf("Write complete code strings and comments to output? (Y/N):");
	scanf("%c%c",&sinput[1],&sinput[0]);  // first %c traps the \n from previous input
	if (('n' == sinput[0]) || ('N' == sinput[0])) doFull = FALSE;
	else doFull = TRUE;

	printf("Write output at HTML table? (Y/N):");
	scanf("%c%c",&sinput[1],&sinput[0]);  // first %c traps the \n from previous input
	if (('n' == sinput[0]) || ('N' == sinput[0])) doHTML = FALSE;
	else doHTML = TRUE;

//printf("Finished\n");
//exit(0);
#else							// get names from macros (debugging)
	strcpy(sfin_name,IN_FILE);
#endif

// open files
	fin = fopen(sfin_name,"r");
	if (!fin) Report_Error("Unable to open input file");
			
	strcpy(sfilt_name,sfin_name);  // create output file name and open file
	if (doHTML) strcat(sfilt_name,HTML_SUFFIX);
	else strcat(sfilt_name,TXT_SUFFIX);
	ffilt = fopen(sfilt_name,"w");
	printf("Index is in the file \"%s\"\n",sfilt_name);
	if (!ffilt) Report_Error("Unable to open index file");

} // GetFileNames

int GetRecord(void)
// read record from fin; returns '/'-delimited string of unique codes
// in scode
// NOTE: This may generates a problem if there is no space between the end of
//       the string and the start of the code. That should not occur in
//       properly-formatted dictionaries, so for the time being consider
//       this a feature
{
char * s = sinput;
char * sc = scode;
char sa[MAX_CODE_LEN];
int k = 0; // number of codes found

	fgets(sinput,MAX_INPUT_LINE,fin);
	if (feof(fin)) return FALSE;
	while (*s) {
		while ((*s) && ('[' != *s)) ++s; // move to start of codes
		if (*s) {
			++s;
			while ((*s) && (*s > ' ') && (']' != *s)) {
				if ('/' == *s) ++k;  // count internal delimiters
				if (':' == *s) {  // replace the verb delimiter with /
					++k;
					*s = '/';
				}
				*sc++ = *s++; // copy code
			}
			while ((*s) && (']' != *s)) s++; // skip other parts
		}
		if ('/' != *(sc-1)) {
			*sc++ = '/'; // delimit codes
			++k;
		}
	}
	*sc = '\0';
	if (scode[0] == '/') scode[0] = '\0'; // no codes found
//	printf("%s\n",scode);
	if (k > 1) { // remove duplicates; code here is a tad subtle...
		char * sp;
		int i;
		sc = scode;
		while (*sc) {
			k = 0;
			while ('/' != *sc) sa[k++] = *sc++; // copy next code into sa
			sa[k++] = *sc++;
			sa[k] = '\0';
//			printf("mk1 -- %s\n",sa);
			while (sp = strstr(sc,sa)) { // deliberate use of assignment
				i = strlen(sa);
				while (*sp) *sp++ = *(sp+i); // overwrite the duplicate
				*sp = '\0';
			}
		}
	}
//	printf("%s\n",scode);
	return TRUE;
} // GetRecord


int AddText(void)
// adds sinput to textBuf; returns index of initial char
{
char * pc = sinput;
char * pt = &textBuf[itextBuf];
int ka = itextBuf; // save this

	if (itextBuf + strlen(sinput) >= MAX_TEXT) Report_Error("Out of memory for storing text");
	if (doFull) {
		while ((*pc) && ('\n' != *pc)) { // check *pc in case of bad file format
			*pt++ = *pc++;
			++itextBuf;
		}
	} else {  // just copy text prior to '['
		while ((*pc) && ('[' != *pc)) { // check *pc in case of bad file format
			*pt++ = *pc++;
			++itextBuf;
		}
		// remove trailing spaces
		while (' ' == *(pc-1)) {
			--pc;
			--itextBuf;
		}
	}
	textBuf[itextBuf] = '\0';
	++itextBuf;
	return ka;
} // AddText

int AddVerb(void)
// adds sinput to textBuf; returns index of initial char
{
char * pc = sinput;
char * pv = sverb;
char * pt = &textBuf[itextBuf];
int ka = itextBuf; // save this

	if (itextBuf + strlen(sinput) >= MAX_TEXT) Report_Error("Out of memory for storing text");
	if ('-' == sinput[0]) {
		while ((*pc) && ('[' != *pc)) { // check *pc in case of bad file format
			*pt++ = *pc++;
			++itextBuf;
			if ('*' == *(pc-1)) {
				while (*pv) { // insert verb
					*pt++ = *pv++;
					++itextBuf;
				}
			}
		}
		while (' ' == *(pc-1)) {
			--pc;
			--itextBuf;
		}

	}
	else {
		while ((*pc) && (' ' != *pc)) { // check *pc in case of bad file format
			*pt++ = *pc;
			*pv++ = *pc++;
			++itextBuf;
		};
		*pv = '\0';
		while ((*pc) && ('[' != *pc)) pc++;
	}
	
	if (doFull) {
		while ((*pc) && ('\n' != *pc)) { // check *pc in case of bad file format
			*pt++ = *pc++;
			++itextBuf;
		}
	} 
	textBuf[itextBuf] = '\0';
	++itextBuf;
	return ka;
} // AddText

void AddTextRec(struct codeListStruct * pcLS)
{
	pcLS->ptext = malloc(sizeof(struct textListStruct));1;
	if (NULL == pcLS->ptext) Report_Error("Insufficient memory for new text index");
	pcLS->ptext->itext = icurtext;
	pcLS->ptext->pnext = NULL;
} // AddText

void AddRecord(void)
// Adds record to index for code src
{
char * pc;
char * pca;
int result;
struct codeListStruct * pcLS,* ppcLS, * psrc, * pn; 
struct textListStruct * ptLS,* pptLS, * ptar;

//	printf("AR Mk1:%s\n",src);
	ppcLS = codeListHead;
	pcLS = codeListHead;
	while (pcLS && (result = strcmp(src,pcLS->code))) { // traverse list looking for src
		if (result < 0) break; // so insert before pcLS
		ppcLS = pcLS;
		pcLS = pcLS->pnext;
	}

	if (!pcLS) { // got to end of list without finding code; make a new entry
		ppcLS->pnext = malloc(sizeof(struct codeListStruct));
		if (NULL == ppcLS->pnext) Report_Error("Insufficient memory for new source");
		pcLS = ppcLS->pnext;
		strcpy(pcLS->code,src);
		AddTextRec(pcLS);
	} 
	else if (result < 0) { // insert a new entry at ppcLS
		pn = malloc(sizeof(struct codeListStruct));
		if (NULL == pn) Report_Error("Insufficient memory for new source");
		pn->pnext = pcLS;
		ppcLS->pnext = pn;
		strcpy(pn->code,src);
		AddTextRec(pn);
	} 
	else {  // found it; add new text
		ptLS = pcLS->ptext;
		while (ptLS->pnext) ptLS = ptLS->pnext;
		ptLS->pnext = malloc(sizeof(struct textListStruct));1;
		if (NULL == ptLS->pnext) Report_Error("Insufficient memory for new text index");
		ptLS->pnext->itext = icurtext;
		ptLS->pnext->pnext = NULL;
	}
//	printf("AR Mk2\n",src);
	return;
} // AddRecord


void WriteIndex(void)
// display list with identation
{	
char spcode[16];
char stabs[3];
int len = 3;
struct codeListStruct * pc = codeListHead;
struct textListStruct * pt;  // next list item
time_t curtime;

	fprintf(ffilt,"Index for %s\n",sfin_name);
	time(&curtime);
	fprintf(ffilt,"Index generated: %s\n",ctime(&curtime));	

	strcpy(spcode,"$$$");  // initialize to something that shouldn't match
	if (!pc->ptext->pnext) pc = pc->pnext; // skip first entry if empty except for description
	printf("\n\nWriting index:\n");
	while (pc) {
		if(isActor) {		// set up indentation for CAMEO actor files
			if ((len >= 6) && (!strncmp(spcode, pc->code,6))) strcpy(stabs,"\t\t");
			else if ((len >= 3) && (!strncmp(spcode, pc->code,3))) strcpy(stabs,"\t");
			else stabs[0] = '\0';
		}
		else {		// set up indentation for CAMEO/WEIS verb files
			if ((len >= 3) && (!strncmp(spcode, pc->code,3))) strcpy(stabs,"\t\t");
			else if ((len >= 2) && (!strncmp(spcode, pc->code,2))) strcpy(stabs,"\t");
			else stabs[0] = '\0';
		}
		fprintf(ffilt,"\n%s*%s*\n",stabs,pc->code);
		printf("%s\n",pc->code);
		pt = pc->ptext;
		while (pt) {
			fprintf(ffilt,"%s\t%s\n",stabs,&textBuf[pt->itext]);
			pt = pt->pnext;
		}		
		strcpy(spcode, pc->code);
		len = strlen(spcode);
		pc = pc->pnext;
	}
} // WriteIndex


void WriteTable(void)
// display list as HTML table
{	
char spcode[16];
char stabs[32];
int len = 3;
struct codeListStruct * pc = codeListHead;
struct textListStruct * pt;  // next list item
time_t curtime;

// write HTML header
fprintf(ffilt,"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"\n");
fprintf(ffilt,"        \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n");
fprintf(ffilt,"<html xmlns=\"http://www.w3.org/1999/xhtml\">\n");
fprintf(ffilt,"<head>\n\t<title>%s</title>\n</head>\n<body>\n",sfin_name);

// writie identifying information
fprintf(ffilt,"<H1>Index for %s</H1>\n",sfin_name);
time(&curtime);
fprintf(ffilt,"<P>Index generated: %s</P>\n",ctime(&curtime));	

fprintf(ffilt,"<table border = \"1\">\n");

	strcpy(spcode,"$$$");  // initialize to something that shouldn't match
	if (!pc->ptext->pnext) pc = pc->pnext; // skip first entry if empty except for description
	printf("\n\nWriting index:\n");
	while (pc) {
		if(isActor) {		// set up indentation for CAMEO actor files
			if ((len >= 6) && (!strncmp(spcode, pc->code,6))) strcpy(stabs,"<td> </td> <td> </td> ");
			else if ((len >= 3) && (!strncmp(spcode, pc->code,3))) strcpy(stabs,"<td> </td> ");
			else stabs[0] = '\0';
		}
		else {		// set up indentation for CAMEO/WEIS verb files
			if ((len >= 4) && (!strncmp(spcode, pc->code,4))) strcpy(stabs,"<td> </td> <td> </td> ");
			else if ((len >= 3) && (!strncmp(spcode, pc->code,3))) strcpy(stabs,"<td> </td> ");
			else stabs[0] = '\0';
		}
		if (*((pc->code)+2) <= '0') fprintf(ffilt,"<tr>\n%s <td><H2>%s</H2><br>\n",stabs,pc->code); // highlight  event cue codes
		else fprintf(ffilt,"<tr>\n%s <td><b>%s</b><br>\n",stabs,pc->code);
		printf("%s\n",pc->code);
		pt = pc->ptext;
		while (pt) {
			char * pta = &textBuf[pt->itext]; // pointers used to handle word-wrapping
			char * pts = pta;                 //  breaking at '_' and ' '
			char * pte = pta;
			int ka = 0;
			fprintf(ffilt,"&nbsp;&nbsp;&nbsp;");
			while (*pta) {
				while (*pta && (ka < WORDWRAP)) {
					if ('_' == *pta || ' ' == *pta) pte = pta; // set allowable ending
					pta++;
					ka++;
				}
				if (!*pta) pte = pta - 1;
				while (pts <= pte) fputc(*pts++,ffilt); // write from start to end, resetting start
				fprintf(ffilt,"<br>\n");
				if (*pta) fprintf(ffilt,"&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"); // indent
				ka = pta - pte;  // set ka to the number of unprinted chars
			}
			pt = pt->pnext;
		}		
		fprintf(ffilt,"</td>\n</tr>\n");
		strcpy(spcode, pc->code);
		len = strlen(spcode);
		pc = pc->pnext;
	}
fprintf(ffilt,"</table>\n");
fprintf(ffilt,"</body>\n</html>\n"); // terminate HTML 

} // WriteTable


void Report_Error(char s[])  
// exit on early end-of-file without recovery
{
int ka;

	printf("\a\nUnrecoverable Error:%s\nLast line read was\nCodes:%s\n",s,sinput,scode);
	printf("Press <return> to exit");
	ka = getchar();
	exit(0);
} // Error_Exit

